@@ -8,6 +8,8 @@ use quote::format_ident;
 use quote::quote;
 use quote::quote_spanned;
 use quote::ToTokens;
+use syn::parenthesized;
+use syn::parse::Parse;
 use syn::parse_macro_input;
 use syn::spanned::Spanned;
 use syn::token::Brace;
@@ -16,16 +18,20 @@ use syn::token::Lt;
 use syn::token::Trait;
 use syn::AngleBracketedGenericArguments;
 use syn::Block;
+use syn::Error;
 use syn::Expr;
 use syn::GenericArgument;
 use syn::GenericParam;
+use syn::Ident;
 use syn::ImplItemMethod;
 use syn::ItemTrait;
+use syn::LitInt;
 use syn::Path;
 use syn::PathArguments;
 use syn::PathSegment;
 use syn::Signature;
 use syn::Stmt;
+use syn::Token;
 use syn::TraitBound;
 use syn::TraitItem;
 use syn::TraitItemMethod;
@@ -37,14 +43,15 @@ use syn::Visibility;
 use syn_utils::TypeOrPath;
 
 use crate::parse_assoc_type::parse_assoc_type;
-use crate::parse_assoc_type::AssocTypeParseError;
+use crate::parse_assoc_type::AssocTypeError;
 use crate::parse_trait_sig::parse_trait_signature;
-use crate::parse_trait_sig::MethodParseError;
+use crate::parse_trait_sig::MethodError;
 use crate::parse_trait_sig::SignatureChanges;
 use crate::parse_trait_sig::TypeTransform;
 use crate::syn_utils::iter_path;
 use crate::syn_utils::trait_bounds;
-use crate::transform::AssocTypeConversions;
+use crate::transform::TransformError;
+use crate::transform::TypeConverter;
 
 mod parse_assoc_type;
 mod parse_trait_sig;
@@ -52,9 +59,35 @@ mod syn_utils;
 mod transform;
 
 macro_rules! abort {
-    ($span:expr, $message:literal) => {
-        quote_spanned! {$span => compile_error!($message);}.into()
-    };
+    ($span:expr, $message:literal $(,$args:expr)*) => {{
+        let msg = format!($message $(,$args)*);
+        let tokens = quote_spanned! {$span => compile_error!(#msg);}.into();
+        tokens
+    }};
+}
+
+struct Collection {
+    id: Ident,
+    count: usize,
+}
+
+impl Parse for Collection {
+    fn parse(input: syn::parse::ParseStream) -> syn::Result<Self> {
+        let inner;
+        parenthesized!(inner in input);
+
+        let id: Ident = inner.parse()?;
+        let _: Token![,] = inner.parse()?;
+        let count_lit: LitInt = inner.parse()?;
+        let count: usize = count_lit.base10_parse()?;
+        if count < 1 {
+            return Err(Error::new(
+                count_lit.span(),
+                "number of type parameters must be >= 1",
+            ));
+        }
+        Ok(Self { id, count })
+    }
 }
 
 #[proc_macro_attribute]
@@ -66,67 +99,7 @@ pub fn dynamize(_attr: TokenStream, input: TokenStream) -> TokenStream {
 
     let mut objectifiable_methods: Vec<(Signature, SignatureChanges)> = Vec::new();
 
-    let mut assoc_type_conversions = AssocTypeConversions::default();
-
-    for item in &original_trait.items {
-        if let TraitItem::Type(assoc_type) = item {
-            match parse_assoc_type(assoc_type) {
-                Err((_, AssocTypeParseError::NoIntoBound)) => continue,
-                Err((span, AssocTypeParseError::AssocTypeInBound)) => {
-                    return abort!(span, "dynamize does not support associated types here")
-                }
-                Err((span, AssocTypeParseError::GenericAssociatedType)) => {
-                    return abort!(
-                        span,
-                        "dynamize does not (yet?) support generic associated types"
-                    )
-                }
-                Ok((ident, type_)) => {
-                    assoc_type_conversions.0.insert(ident.clone(), type_);
-                }
-            }
-        }
-    }
-
-    for item in &original_trait.items {
-        if let TraitItem::Method(method) = item {
-            let mut signature = method.sig.clone();
-            match parse_trait_signature(&mut signature, &assoc_type_conversions) {
-                Err((_, MethodParseError::NonDispatchableMethod)) => continue,
-                Err((span, MethodParseError::UnconvertibleAssocType)) => {
-                    return abort!(
-                        span,
-                        "associated type is either undefined or doesn't have a trait bound"
-                    )
-                }
-                Err((span, MethodParseError::AssocTypeInInputs)) => {
-                    return abort!(
-                        span,
-                        "dynamize does not support associated types in parameter types"
-                    )
-                }
-                Err((
-                    span,
-                    MethodParseError::AssocTypeInUnsupportedReturnType
-                    | MethodParseError::UnconvertibleAssocTypeInFnInput,
-                )) => return abort!(span, "dynamize does not know how to convert this type"),
-                Err((
-                    span,
-                    MethodParseError::UnconvertibleAssocTypeInTraitBound
-                    | MethodParseError::UnconvertibleAssocTypeInWhereClause,
-                )) => return abort!(span, "dynamize does not support associated types here"),
-                Err((span, MethodParseError::ImplTraitInInputs)) => {
-                    return abort!(
-                        span,
-                        "dynamize does not support impl here, change it to a method generic"
-                    )
-                }
-                Ok(parsed_method) => objectifiable_methods.push((signature, parsed_method)),
-            };
-        }
-    }
-
-    let mut method_impls: Vec<ImplItemMethod> = Vec::new();
+    let mut type_converter = TypeConverter::default();
 
     let mut blanket_impl_attrs = Vec::new();
     let mut dyn_trait_attrs = Vec::new();
@@ -157,11 +130,102 @@ pub fn dynamize(_attr: TokenStream, input: TokenStream) -> TokenStream {
             };
             let tokens = group.stream();
             dyn_trait_attrs.push(quote! {#[#tokens]});
+        } else if original_trait.attrs[i].path.is_ident("collection") {
+            let attr = original_trait.attrs.remove(i);
+            let tokens = attr.tokens.into();
+            let coll = parse_macro_input!(tokens as Collection);
+
+            if type_converter
+                .collections
+                .insert(coll.id.clone(), coll.count)
+                .is_some()
+            {
+                return abort!(
+                    coll.id.span(),
+                    "collection `{}` is defined multiple times for this trait",
+                    coll.id
+                );
+            }
         } else {
             i += 1;
         }
     }
 
+    for item in &original_trait.items {
+        if let TraitItem::Type(assoc_type) = item {
+            match parse_assoc_type(assoc_type) {
+                Ok((ident, type_)) => {
+                    type_converter
+                        .assoc_type_conversions
+                        .insert(ident.clone(), type_);
+                }
+                Err((_, AssocTypeError::NoTraitBound)) => continue,
+                Err((span, AssocTypeError::TraitBoundContainsAssocType)) => {
+                    return abort!(span, "dynamize does not support associated types here")
+                }
+                Err((span, AssocTypeError::GenericAssociatedType)) => {
+                    return abort!(
+                        span,
+                        "dynamize does not (yet?) support generic associated types"
+                    )
+                }
+            }
+        }
+    }
+
+    for item in &original_trait.items {
+        if let TraitItem::Method(method) = item {
+            let mut signature = method.sig.clone();
+            match parse_trait_signature(&mut signature, &type_converter) {
+                Ok(parsed_method) => objectifiable_methods.push((signature, parsed_method)),
+                Err((span, err)) => match err {
+                    MethodError::NonDispatchableMethod => continue,
+                    MethodError::AssocTypeInInputs => {
+                        return abort!(
+                            span,
+                            "dynamize does not support associated types in parameter types"
+                        )
+                    }
+                    MethodError::ImplTraitInInputs => {
+                        return abort!(
+                            span,
+                            "dynamize does not support impl here, change it to a method generic"
+                        )
+                    }
+                    MethodError::Transform(TransformError::AssocTypeWithoutDestType) => {
+                        return abort!(
+                            span,
+                            "associated type is either undefined or doesn't have a trait bound"
+                        )
+                    }
+                    MethodError::Transform(TransformError::UnsupportedType) => {
+                        return abort!(span, "dynamize does not know how to convert this type")
+                    }
+                    MethodError::Transform(TransformError::ExpectedAtLeastNTypes(n)) => {
+                        return abort!(
+                            span,
+                            "dynamize expects at least {} generic type arguments for this type",
+                            n
+                        )
+                    }
+                    MethodError::Transform(TransformError::AssocTypeAfterFirstNTypes(n, ident)) => {
+                        return abort!(
+                            span,
+                            "for {} dynamize supports associated types only within the first {} generic type parameters",
+                            ident,
+                            n
+                        )
+                    }
+                    MethodError::UnconvertedAssocType => {
+                        return abort!(span, "dynamize does not support associated types here")
+                    }
+                },
+            };
+        }
+    }
+
+    let mut method_impls: Vec<ImplItemMethod> = Vec::new();
+
     let mut dyn_trait = ItemTrait {
         ident: format_ident!("Dyn{}", original_trait.ident),
 
@@ -362,7 +426,7 @@ fn generate_blanket_impl(
 }
 
 fn path_is_assoc_type(path: &Path) -> bool {
-    path.segments.first().unwrap().ident == "Self"
+    path.segments[0].ident == "Self"
 }
 
 fn match_assoc_type(item: TypeOrPath) -> bool {
@@ -409,7 +473,16 @@ impl TypeTransform {
                     quote! {#arg.map_err(|x| #err_inner)}
                 }
             }
-            _other => arg,
+            TypeTransform::IntoIterMapCollect(types) => {
+                let idents = (0..types.len()).map(|i| format_ident!("v{}", i));
+                // FUTURE: let transforms = std::iter::zip(idents, types).map(|(i, t)| t.convert(quote! {#i}));
+                let transforms = types.iter().enumerate().map(|(idx, t)| {
+                    let id = format_ident!("v{}", idx);
+                    t.convert(quote! {#id})
+                });
+                quote! {#arg.into_iter().map(|(#(#idents),*)| (#(#transforms),*)).collect()}
+            }
+            TypeTransform::NoOp => arg,
         }
     }
 }
