@@ -8,10 +8,10 @@ use crate::parse_trait_sig::TypeTransform;
 use crate::syn_utils::{iter_type, lifetime_bounds, trait_bounds};
 
 #[derive(Debug)]
-pub enum AssocTypeParseError {
-    AssocTypeInBound,
+pub enum AssocTypeError {
+    NoTraitBound,
+    TraitBoundContainsAssocType,
     GenericAssociatedType,
-    NoIntoBound,
 }
 
 #[derive(Debug, Clone)]
@@ -53,25 +53,28 @@ impl DestType<'_> {
 
 pub fn parse_assoc_type(
     assoc_type: &TraitItemType,
-) -> Result<(&Ident, DestType), (Span, AssocTypeParseError)> {
+) -> Result<(&Ident, DestType), (Span, AssocTypeError)> {
     if let Some(bound) = trait_bounds(&assoc_type.bounds).next() {
         if let PathSegment {
             ident,
             arguments: PathArguments::AngleBracketed(args),
-        } = bound.path.segments.first().unwrap()
+        } = &bound.path.segments[0]
         {
             if ident == "Into" && args.args.len() == 1 {
-                if let GenericArgument::Type(into_type) = args.args.first().unwrap() {
+                if let GenericArgument::Type(into_type) = &args.args[0] {
                     // provide a better error message for type A: Into<Self::B>
                     if iter_type(into_type).any(match_assoc_type) {
-                        return Err((into_type.span(), AssocTypeParseError::AssocTypeInBound));
+                        return Err((
+                            into_type.span(),
+                            AssocTypeError::TraitBoundContainsAssocType,
+                        ));
                     }
 
                     // TODO: support lifetime GATs (see the currently failing tests/gats.rs)
                     if !assoc_type.generics.params.is_empty() {
                         return Err((
                             assoc_type.generics.params.span(),
-                            AssocTypeParseError::GenericAssociatedType,
+                            AssocTypeError::GenericAssociatedType,
                         ));
                     }
 
@@ -89,7 +92,7 @@ pub fn parse_assoc_type(
             }),
         ));
     }
-    Err((assoc_type.span(), AssocTypeParseError::NoIntoBound))
+    Err((assoc_type.span(), AssocTypeError::NoTraitBound))
 }
 
 #[cfg(test)]
@@ -97,7 +100,7 @@ mod tests {
     use quote::quote;
     use syn::{TraitItemType, Type};
 
-    use crate::parse_assoc_type::{parse_assoc_type, AssocTypeParseError, DestType};
+    use crate::parse_assoc_type::{parse_assoc_type, AssocTypeError, DestType};
 
     #[test]
     fn ok() {
@@ -122,7 +125,7 @@ mod tests {
 
         assert!(matches!(
             parse_assoc_type(&type1),
-            Err((_, AssocTypeParseError::NoIntoBound))
+            Err((_, AssocTypeError::NoTraitBound))
         ));
     }
 
@@ -135,7 +138,7 @@ mod tests {
 
         assert!(matches!(
             parse_assoc_type(&type1),
-            Err((_, AssocTypeParseError::AssocTypeInBound))
+            Err((_, AssocTypeError::TraitBoundContainsAssocType))
         ));
     }
 
@@ -148,7 +151,7 @@ mod tests {
 
         assert!(matches!(
             parse_assoc_type(&type1),
-            Err((_, AssocTypeParseError::GenericAssociatedType))
+            Err((_, AssocTypeError::GenericAssociatedType))
         ));
     }
 
@@ -161,7 +164,7 @@ mod tests {
 
         assert!(matches!(
             parse_assoc_type(&type1),
-            Err((_, AssocTypeParseError::GenericAssociatedType))
+            Err((_, AssocTypeError::GenericAssociatedType))
         ));
     }
 }
