@@ -10,24 +10,56 @@ use syn::{
 use crate::{
     filter_map_assoc_paths, match_assoc_type,
     parse_assoc_type::{BoxType, DestType},
-    parse_trait_sig::{MethodParseError, TypeTransform},
-    syn_utils::{iter_path, iter_type},
+    parse_trait_sig::{MethodError, TypeTransform},
+    syn_utils::{iter_path, iter_type, type_arguments_mut},
 };
 
 #[derive(Default)]
-pub struct AssocTypeConversions<'a>(pub HashMap<Ident, DestType<'a>>);
-
-pub enum TransformError {
-    UnconvertibleAssocType(Span),
-    AssocTypeInUnsupportedType(Span),
+pub struct TypeConverter<'a> {
+    pub assoc_type_conversions: HashMap<Ident, DestType<'a>>,
+    pub collections: HashMap<Ident, usize>,
 }
 
-impl AssocTypeConversions<'_> {
-    pub fn parse_type_path(&self, type_: &mut Type) -> Result<TypeTransform, TransformError> {
-        let assoc_span = match iter_type(type_).filter_map(filter_map_assoc_paths).next() {
-            Some(path) => path.span(),
-            None => return Ok(TypeTransform::NoOp),
-        };
+#[derive(Debug)]
+pub enum TransformError {
+    AssocTypeWithoutDestType,
+    UnsupportedType,
+    ExpectedAtLeastNTypes(usize),
+    AssocTypeAfterFirstNTypes(usize, Ident),
+}
+
+impl TypeConverter<'_> {
+    /// A return type of Some(1) means that the type implements
+    /// IntoIterator<Item=T1> and FromIterator<T1>
+    /// with T1 being its first generic type parameter.
+    ///
+    /// A return type of Some(2) means that the type implements
+    /// IntoIterator<Item=(T1, T2)> and FromIterator<(T1, T2)>
+    /// with T1 and T2 being its first two generic type parameters.
+    ///
+    /// ... etc. A return type of None means the type isn't recognized.
+    #[rustfmt::skip]
+    fn get_collection_type_count(&self, ident: &Ident) -> Option<usize> {
+        if let Some(count) = self.collections.get(ident) {
+            return Some(*count);
+        }
+
+        // when adding a type here don't forget to document it in the README
+        if ident == "Vec"        { return Some(1); }
+        if ident == "VecDeque"   { return Some(1); }
+        if ident == "LinkedList" { return Some(1); }
+        if ident == "HashSet"    { return Some(1); }
+        if ident == "BinaryHeap" { return Some(1); }
+        if ident == "BTreeSet"   { return Some(1); }
+        if ident == "HashMap"    { return Some(2); }
+        if ident == "BTreeMap"   { return Some(2); }
+        None
+    }
+
+    pub fn convert_type(&self, type_: &mut Type) -> Result<TypeTransform, (Span, TransformError)> {
+        if !iter_type(type_).any(match_assoc_type) {
+            return Ok(TypeTransform::NoOp);
+        }
 
         if let Type::Reference(TypeReference {
             lifetime: None,
@@ -42,20 +74,18 @@ impl AssocTypeConversions<'_> {
             }) = elem.as_mut()
             {
                 if bounds.len() == 1 {
-                    if let TypeParamBound::Trait(bound) = bounds.first_mut().unwrap() {
+                    if let TypeParamBound::Trait(bound) = &mut bounds[0] {
                         if bound.path.segments.len() == 1 {
-                            let first = bound.path.segments.first_mut().unwrap();
+                            let first = &mut bound.path.segments[0];
                             if first.ident == "Iterator" {
                                 if let PathArguments::AngleBracketed(args) = &mut first.arguments {
                                     if args.args.len() == 1 {
-                                        if let GenericArgument::Binding(binding) =
-                                            args.args.first_mut().unwrap()
+                                        if let GenericArgument::Binding(binding) = &mut args.args[0]
                                         {
                                             if binding.ident == "Item"
                                                 && iter_type(&binding.ty).any(match_assoc_type)
                                             {
-                                                let inner =
-                                                    self.parse_type_path(&mut binding.ty)?;
+                                                let inner = self.convert_type(&mut binding.ty)?;
                                                 let box_type = BoxType {
                                                     inner: quote! {#elem},
                                                     placeholder_lifetime: true,
@@ -77,81 +107,82 @@ impl AssocTypeConversions<'_> {
         }
 
         if let Type::Path(TypePath { path, qself: None }) = type_ {
-            let ident = &path.segments.first().unwrap().ident;
-
-            if ident == "Self" && path.segments.len() == 2 {
-                let ident = &path.segments.last().unwrap().ident;
-                let dest_type = self
-                    .0
-                    .get(ident)
-                    .ok_or_else(|| TransformError::UnconvertibleAssocType(ident.span()))?;
-                *type_ = dest_type.get_dest();
-                return Ok(dest_type.type_transformation());
-            } else if ident == "Option" && path.segments.len() == 1 {
-                let first_seg = path.segments.first_mut().unwrap();
-
-                if let PathArguments::AngleBracketed(args) = &mut first_seg.arguments {
-                    if args.args.len() == 1 {
-                        if let GenericArgument::Type(generic_type) = args.args.first_mut().unwrap()
-                        {
-                            if iter_type(generic_type).any(match_assoc_type) {
-                                return Ok(TypeTransform::Map(
-                                    self.parse_type_path(generic_type)?.into(),
-                                ));
-                            }
-                        }
-                    }
-                }
-            } else if ident == "Result" && path.segments.len() == 1 {
-                let first_seg = path.segments.first_mut().unwrap();
-                if let PathArguments::AngleBracketed(args) = &mut first_seg.arguments {
-                    if args.args.len() == 2 {
-                        let mut args_iter = args.args.iter_mut();
-                        if let (GenericArgument::Type(ok_type), GenericArgument::Type(err_type)) =
-                            (args_iter.next().unwrap(), args_iter.next().unwrap())
-                        {
-                            if iter_type(ok_type).any(match_assoc_type)
-                                || iter_type(err_type).any(match_assoc_type)
-                            {
-                                return Ok(TypeTransform::Result(
-                                    self.parse_type_path(ok_type)?.into(),
-                                    self.parse_type_path(err_type)?.into(),
-                                ));
-                            }
-                        }
-                    }
+            if path.segments[0].ident == "Self" {
+                if path.segments.len() == 2 {
+                    let ident = &path.segments.last().unwrap().ident;
+                    let dest_type = self
+                        .assoc_type_conversions
+                        .get(ident)
+                        .ok_or_else(|| (ident.span(), TransformError::AssocTypeWithoutDestType))?;
+                    *type_ = dest_type.get_dest();
+                    return Ok(dest_type.type_transformation());
                 }
             } else {
-                let last_seg = &path.segments.last().unwrap();
-                if last_seg.ident == "Result" {
-                    let last_seg = path.segments.last_mut().unwrap();
-                    if let PathArguments::AngleBracketed(args) = &mut last_seg.arguments {
-                        if args.args.len() == 1 {
-                            if let GenericArgument::Type(generic_type) =
-                                args.args.first_mut().unwrap()
-                            {
-                                if iter_type(generic_type).any(match_assoc_type) {
-                                    return Ok(TypeTransform::Map(
-                                        self.parse_type_path(generic_type)?.into(),
+                let path_len = path.segments.len();
+                let last_seg = path.segments.last_mut().unwrap();
+
+                if let PathArguments::AngleBracketed(args) = &mut last_seg.arguments {
+                    let mut args: Vec<_> = type_arguments_mut(&mut args.args).collect();
+
+                    if path_len == 1 {
+                        if let Some(type_count) = self.get_collection_type_count(&last_seg.ident) {
+                            if args.len() < type_count {
+                                return Err((
+                                    last_seg.span(),
+                                    TransformError::ExpectedAtLeastNTypes(type_count),
+                                ));
+                            }
+                            for i in type_count..args.len() {
+                                if iter_type(args[i]).any(match_assoc_type) {
+                                    return Err((
+                                        args[i].span(),
+                                        TransformError::AssocTypeAfterFirstNTypes(
+                                            type_count,
+                                            last_seg.ident.clone(),
+                                        ),
                                     ));
                                 }
                             }
+                            let mut transforms = Vec::new();
+                            for arg in args {
+                                transforms.push(self.convert_type(arg)?);
+                            }
+                            return Ok(TypeTransform::IntoIterMapCollect(transforms));
                         }
                     }
+
+                    if args.len() == 1 {
+                        if iter_type(args[0]).any(match_assoc_type)
+                            && ((last_seg.ident == "Option" && path_len == 1)
+                                || last_seg.ident == "Result")
+                        {
+                            return Ok(TypeTransform::Map(self.convert_type(args[0])?.into()));
+                        }
+                    } else if args.len() == 2
+                        && path_len == 1
+                        && (iter_type(args[0]).any(match_assoc_type)
+                            || iter_type(args[1]).any(match_assoc_type))
+                        && last_seg.ident == "Result"
+                    {
+                        return Ok(TypeTransform::Result(
+                            self.convert_type(args[0])?.into(),
+                            self.convert_type(args[1])?.into(),
+                        ));
+                    }
                 }
             }
         }
 
         // the type contains an associated type but we
         // don't know how to deal with it so we abort
-        Err(TransformError::AssocTypeInUnsupportedType(assoc_span))
+        Err((type_.span(), TransformError::UnsupportedType))
     }
 }
 
 pub fn dynamize_function_bounds(
     generics: &mut Generics,
-    assoc_type_conversions: &AssocTypeConversions,
-) -> Result<HashMap<Ident, Vec<TypeTransform>>, (Span, MethodParseError)> {
+    type_converter: &TypeConverter,
+) -> Result<HashMap<Ident, Vec<TypeTransform>>, (Span, MethodError)> {
     let mut type_param_transforms = HashMap::new();
 
     for type_param in generics.type_params_mut() {
@@ -159,7 +190,7 @@ pub fn dynamize_function_bounds(
             if let TypeParamBound::Trait(bound) = bound {
                 dynamize_trait_bound(
                     bound,
-                    assoc_type_conversions,
+                    type_converter,
                     &type_param.ident,
                     &mut type_param_transforms,
                 )?;
@@ -176,7 +207,7 @@ pub fn dynamize_function_bounds(
                             if let TypeParamBound::Trait(bound) = bound {
                                 dynamize_trait_bound(
                                     bound,
-                                    assoc_type_conversions,
+                                    type_converter,
                                     ident,
                                     &mut type_param_transforms,
                                 )?;
@@ -190,10 +221,7 @@ pub fn dynamize_function_bounds(
                 if let Some(assoc_type) =
                     iter_type(&predicate_type.bounded_ty).find_map(filter_map_assoc_paths)
                 {
-                    return Err((
-                        assoc_type.span(),
-                        MethodParseError::UnconvertibleAssocTypeInWhereClause,
-                    ));
+                    return Err((assoc_type.span(), MethodError::UnconvertedAssocType));
                 }
 
                 // just to provide better error messages
@@ -202,10 +230,7 @@ pub fn dynamize_function_bounds(
                         if let Some(assoc_type) =
                             iter_path(&bound.path).find_map(filter_map_assoc_paths)
                         {
-                            return Err((
-                                assoc_type.span(),
-                                MethodParseError::UnconvertibleAssocTypeInWhereClause,
-                            ));
+                            return Err((assoc_type.span(), MethodError::UnconvertedAssocType));
                         }
                     }
                 }
@@ -218,26 +243,23 @@ pub fn dynamize_function_bounds(
 
 fn dynamize_trait_bound(
     bound: &mut TraitBound,
-    assoc_type_conversions: &AssocTypeConversions,
+    type_converter: &TypeConverter,
     type_ident: &Ident,
     type_param_transforms: &mut HashMap<Ident, Vec<TypeTransform>>,
-) -> Result<(), (Span, MethodParseError)> {
+) -> Result<(), (Span, MethodError)> {
     if bound.path.segments.len() == 1 {
-        let segment = bound.path.segments.first_mut().unwrap();
+        let segment = &mut bound.path.segments[0];
 
         if let PathArguments::Parenthesized(args) = &mut segment.arguments {
             if segment.ident == "Fn" || segment.ident == "FnOnce" || segment.ident == "FnMut" {
                 let mut transforms = Vec::new();
                 for input_type in &mut args.inputs {
-                    match assoc_type_conversions.parse_type_path(input_type) {
+                    match type_converter.convert_type(input_type) {
                         Ok(ret_type) => {
                             transforms.push(ret_type);
                         }
-                        Err(TransformError::UnconvertibleAssocType(span)) => {
-                            return Err((span, MethodParseError::UnconvertibleAssocType));
-                        }
-                        Err(TransformError::AssocTypeInUnsupportedType(span)) => {
-                            return Err((span, MethodParseError::UnconvertibleAssocTypeInFnInput));
+                        Err((span, err)) => {
+                            return Err((span, err.into()));
                         }
                     }
                 }
@@ -251,10 +273,7 @@ fn dynamize_trait_bound(
         .filter_map(filter_map_assoc_paths)
         .next()
     {
-        return Err((
-            path.span(),
-            MethodParseError::UnconvertibleAssocTypeInTraitBound,
-        ));
+        return Err((path.span(), MethodError::UnconvertedAssocType));
     }
     Ok(())
 }
