@@ -204,6 +204,36 @@ fn test3<T: TypeWithSuper>(some: T) {
     println!("{}", dyn_trait);
 }
 
+use std::collections::{BTreeMap, BTreeSet, BinaryHeap, HashMap, HashSet, LinkedList, VecDeque};
+
+#[dynamize::dynamize]
+trait Collections {
+    type A: std::error::Error;
+
+    fn vec(&self) -> Vec<Self::A>;
+
+    fn vec_deque(&self) -> VecDeque<Self::A>;
+
+    fn linked_list(&self) -> LinkedList<Self::A>;
+
+    fn vec_opt(&self) -> Vec<Option<Self::A>>;
+}
+
+#[dynamize::dynamize]
+trait CollectionsRequiringEq {
+    type A: Into<String>;
+
+    fn hashset(&self) -> HashSet<Self::A>;
+
+    fn binary_heap(&self) -> BinaryHeap<Self::A>;
+
+    fn btree_set(&self) -> BTreeSet<Self::A>;
+
+    fn hash_map(&self) -> HashMap<Self::A, Self::A>;
+
+    fn btree_map(&self) -> BTreeMap<Self::A, Self::A>;
+}
+
 #[dynamize::dynamize]
 trait ReturnIter {
     type A: std::error::Error;
@@ -221,3 +251,28 @@ trait FunIter {
 
     fn foobar2<H: Fn(&mut dyn Iterator<Item = &mut dyn Iterator<Item = Self::A>>)>(&mut self, f: H);
 }
+
+struct MyCollection<A, B, C>(A, B, C);
+impl<A, B, C> IntoIterator for MyCollection<A, B, C> {
+    type Item = (A, B, C);
+
+    type IntoIter = Box<dyn Iterator<Item = (A, B, C)>>;
+
+    fn into_iter(self) -> Self::IntoIter {
+        todo!()
+    }
+}
+
+impl<A, B, C> FromIterator<(A, B, C)> for MyCollection<A, B, C> {
+    fn from_iter<T: IntoIterator<Item = (A, B, C)>>(_iter: T) -> Self {
+        todo!()
+    }
+}
+
+#[dynamize::dynamize]
+#[collection(MyCollection, 3)]
+trait CustomCollection {
+    type A: Into<String>;
+
+    fn test(&self) -> MyCollection<Self::A, Self::A, Self::A>;
+}
